/*

File: buffers.c
Version: v0.3 nov 2008

	Copyright (C) 2008
		Eric Giesselbach
		
	This software is provided 'as-is', without any express or implied
	warranty.  In no event will the authors be held liable for any
	damages arising from the use of this software.

	Permission is granted to anyone to use this software for any
	purpose, including commercial applications, and to alter it and
	redistribute it freely, subject to the following restrictions:

	1.	The origin of this software must not be misrepresented; you
		must not claim that you wrote the original software. If you use
		this software in a product, an acknowledgment in the product
		documentation would be appreciated but is not required.
	2.	Altered source versions must be plainly marked as such, and
		must not be misrepresented as being the original software.
	3.	This notice may not be removed or altered from any source
		distribution.

*/

#include <nds.h>
#include <nds/arm9/console.h>
#include <stdio.h>
#include <string.h>
#include "802.11.h"
#include <common.h>
#include "buffers.h"
#include <fastmath.h>

extern volatile unsigned char *sndBufferMirror;
extern volatile u8 *micBufferMirror;

extern u8* wifiOutBuffer;
extern u32 wifiOutBufferLength;
extern u8* wifiInBuffer;
extern u32 wifiInBufferLength;

extern u32 disppwr;
extern u32 dispqty;

extern tPacketHeader *packetHeader;

// number of bytes (samples) in VBlank frame
//static float audioBytesInFrame = (float)SAMPLERATE/VBLANK_FREQ; 
static float audioBytesInFrame = (float)SAMPLERATE/VBLANK_FREQ; 

static int counterdelta = 0;
static int soundoffset = 0;
static u32 remotecounter = 0;
static u32 samplesToRead = 0;

static u32 samplesIn  = 0;
static u32 samplesOut = 0;
static int localAudioDelta = 0;
static int prevLocalAudioDelta = 0;
static u8 sample = 0;

static int test = -1;

static u32 totalMicSamples = 0;

extern int debug;

//                         samplesToRead
//  0                  >-----<                 AUDIOBUFLEN
//  |------------------------------------------|
//                      |    |
//                      |    |__ micWriterPos (last written sample)
//                      |__ micReaderPos
//
//  0                                          AUDIOBUFLEN
//  |------------------------------------------|
//               |    |
//               |    |__ sndWriterPos
//               |__ sndReaderPos (last read sample)
//
//
//  micIndex: absolute mic sample index
//  micCount: actual mic sample count (micIndex - interrupt loss)
//  sndIndex: absoulte audio sample counter
//

u32 sndReaderPos;
u32 sndWriterPos;
u32 sndWriterIndex;

u32 micReaderPos;
u32 micWriterPos;

u32 micIndex = 0;
u32 micCount = 0;
u32 sndIndex = 0;
u32 audioIndex = 0; // mic index to send to other side

// Talkie->micWriterPos is current micBuffer position
// Talkie->sndOffset is micBuffer position on audio out start
// Talkie->micSampleCount is audio (byte) counter
// Talkie->micFrames is number of VBlanks after mic start
// Talkie->micFrameOffset is number of mic samples after last VBlank	

static u32 get_mic_counterVB()
{	
	// get theoretical microphone audio index(when interrupt hasn't been blocked)
	return (int)((float)Talkie->micFrames * audioBytesInFrame + Talkie->micFrameOffset);
}

//prints a nicely formatted debug string, showing the read/write status.
static void print_debug() 
{
	int loss;
	
	if (samplesOut > 0)
		loss = 100 - (int)((100 * samplesIn) / samplesOut);
	
	if (debug == 1)
	{
		printf("\x1b[8;0H");
		printf("mic delta  : %i          \n", totalMicSamples - 4084 );
		printf("samples in : %u          \n", samplesIn);
		printf("packet loss: %i %c       \n", loss, '%');
		printf("X : %u          \n", Talkie->countA);	
		printf("Y : %u          \n", Talkie->countB);	
		printf("wifi tot   : %u          \n", IPC->countC);	
		printf("ARM7 frames: %u          \n", Talkie->countD);	
		printf("ARM7 touch : %u          \n", Talkie->countC);	
		printf("audio delta: %i          \n", localAudioDelta);	
		printf("mic index  : %u          \n", micIndex);
		printf("mic count  : %u          \n", micCount);
		printf("remote index : %u        \n", remotecounter);
		printf("remote delta : %i        \n", counterdelta);
		printf("sound offset : %i        \n", soundoffset);
	}
	
	dispqty = (int)((75 * samplesIn) / samplesOut);		
	if (dispqty > 75) dispqty = 75;
	

	samplesIn = 0;
	samplesOut = 0;
	Talkie->countA = 0;
	Talkie->countB = 0;
	Talkie->countC = 0;
	Talkie->countD = 0;
	
	totalMicSamples = 0;
}

void write_buffer(u32 live) // on VBlanc, live: contact established
{
   u32 wifiInBufferPos;
	
	// microphone counters
	micWriterPos = Talkie->micWriterPos;   // last sample position in micBuffer
	micIndex = get_mic_counterVB();        // theoretical mic sample counter
	micCount = Talkie->micSampleCount;     // real mic sample counter (some samples lost)
	prevLocalAudioDelta = localAudioDelta; 
	localAudioDelta = micCount - micIndex; // < 0 : lost samples (interrupt blocking)
	
	if ( test == -1 )
	{
		micReaderPos = micWriterPos - audioBytesInFrame - 1;	// next input sample position
		if ( micReaderPos < 0 ) micReaderPos += AUDIOBUFLEN;	
	}
	
	// sound counters
	sndIndex = micIndex - Talkie->sndOffset; // played sample counter
	sndReaderPos = sndIndex % AUDIOBUFLEN;   // last played sample position in sndBuffer
	sndWriterIndex = sndIndex + 15 * audioBytesInFrame;  // where to write next input sample (when micCount == micIndex)
	
	// samples to read from mic buffer
	if (micReaderPos <= micWriterPos)
		samplesToRead = micWriterPos - micReaderPos;
	else
		samplesToRead = AUDIOBUFLEN - (micReaderPos - micWriterPos);
	
	// wifiOutBuffer starts with u32 containing audio index
	if ( wifiOutBufferLength == 0 )
	{
		packetHeader = (tPacketHeader*)wifiOutBuffer;
		packetHeader->audioIndex = sndWriterIndex + prevLocalAudioDelta;
		wifiOutBufferLength += sizeof(tPacketHeader);
	}
		
	u32 cnt = 0;
	
	totalMicSamples += samplesToRead;
	
	while ( cnt < samplesToRead )
	{
		// from mic to wifi out
		sample = micBufferMirror[micReaderPos];
		if( ++micReaderPos >= AUDIOBUFLEN) micReaderPos = 0;
							
		wifiOutBuffer[wifiOutBufferLength] = sample;
		samplesOut++;
		
		if (++wifiOutBufferLength >= AUDIOBUFLEN) 
			wifiOutBufferLength = AUDIOBUFLEN -1;
		
		cnt++;
	}
	
	// insert samples to compensate for microphone lag caused by blocked interrupts
	if (localAudioDelta < 0)
	{
		Talkie->micSampleCount++;
		IPC->mailData++;
		Talkie->micSampleCount++;
		IPC->mailData++;
		Talkie->micSampleCount++;
		IPC->mailData++;
		
		micCount++;
		micCount++;
		micCount++;
		
		wifiOutBuffer[wifiOutBufferLength++] = sample;
		if (wifiOutBufferLength >= AUDIOBUFLEN) wifiOutBufferLength = AUDIOBUFLEN -1;
		wifiOutBuffer[wifiOutBufferLength++] = sample;
		if (wifiOutBufferLength >= AUDIOBUFLEN) wifiOutBufferLength = AUDIOBUFLEN -1;
		wifiOutBuffer[wifiOutBufferLength++] = sample;
		if (wifiOutBufferLength >= AUDIOBUFLEN) wifiOutBufferLength = AUDIOBUFLEN -1;
		
		samplesOut++;
		samplesOut++;		
		samplesOut++;
	}
	
		// flush buffer on contact timeout
	if (live == 0)
	{
		samplesOut = 0;
		wifiOutBufferLength = 0;
	}
	
	wifiInBufferPos = 0;
	u32 nextPacketPos;
	
	// get remote audio, possibly multiple packets
	while (wifiInBufferPos < wifiInBufferLength && live == 1)
	{ 
		if ( (wifiInBufferLength - wifiInBufferPos) >= sizeof(tPacketHeader) )
		{
			packetHeader = (tPacketHeader*)&wifiInBuffer[wifiInBufferPos];
			remotecounter = packetHeader->audioIndex;
			wifiInBufferPos += sizeof(tPacketHeader);
			nextPacketPos =  wifiInBufferPos + packetHeader->sampleCount;
			if (nextPacketPos > wifiInBufferLength)
				nextPacketPos = wifiInBufferLength;
			
			if (counterdelta == 0) // first frame
				counterdelta = sndWriterIndex - remotecounter;
			
			// where to place frame relative to own sndWriterIndex?
			sndWriterPos = (counterdelta + remotecounter) % AUDIOBUFLEN;
			if (sndWriterPos < 0) sndWriterPos += AUDIOBUFLEN;
			soundoffset = counterdelta + remotecounter - sndWriterIndex;
			
			// soundoffset < frames ahead: packet outside jitter buffer, advance to next packet
			if (soundoffset < -12 * audioBytesInFrame) wifiInBufferPos += nextPacketPos;
	
			// write input samples to sndBuffer
			while ( wifiInBufferPos < nextPacketPos )
			{
				// from wifi in to sndBuffer
				sample = wifiInBuffer[wifiInBufferPos++];
				samplesIn++;
								
				sndBufferMirror[sndWriterPos] = sample ^ 0x80;
				if( ++sndWriterPos >= AUDIOBUFLEN) sndWriterPos = 0;
			}	
		}
		 else wifiInBufferPos = wifiInBufferLength;
	}
	
	// clear part of played buffer
	sndWriterPos = ( sndIndex - (int)(10 * audioBytesInFrame) + 1 ) % AUDIOBUFLEN;
	if ( sndWriterPos >= -audioBytesInFrame ) 
	{
		cnt = 0;
		while (cnt++ < audioBytesInFrame + 2) sndBufferMirror[sndWriterPos++] = 0;
	}
		
	if (live == 0) // reset counter
		counterdelta = 0;
	
	if( test == -1 || test % 30 == 0 ) // 2x p sec 
		print_debug();

	wifiInBufferLength = 0;
	
	test++;
}

/*
void write_buffer(u32 live) // on VBlanc, live: contact established
{
   u32 wifiInBufferPos;
	
	// microphone counters
	micWriterPos = Talkie->micWriterPos;   // last sample position in micBuffer
	micIndex = get_mic_counterVB();        // theoretical mic sample counter
	micCount = Talkie->micSampleCount;     // real mic sample counter (some samples lost)
	prevLocalAudioDelta = localAudioDelta; 
	localAudioDelta = micCount - micIndex; // < 0 : lost samples (interrupt blocking)
	
	if ( test == -1 )
{
		micReaderPos = micWriterPos - audioBytesInFrame - 1;	// next input sample position
		if ( micReaderPos < 0 ) micReaderPos += AUDIOBUFLEN;	
}
	
	// sound counters
	sndIndex = micIndex - Talkie->sndOffset; // played sample counter
	sndReaderPos = sndIndex % AUDIOBUFLEN;   // last played sample position in sndBuffer
	sndWriterIndex = sndIndex + 15 * audioBytesInFrame;  // where to write next input sample (when micCount == micIndex)
	
	// samples to read from mic buffer
	if (micReaderPos <= micWriterPos)
		samplesToRead = micWriterPos - micReaderPos;
	else
		samplesToRead = AUDIOBUFLEN - (micReaderPos - micWriterPos);
	
	// wifiOutBuffer starts with u32 containing audio index
	if ( wifiOutBufferLength == 0 )
{
		packetHeader = (tPacketHeader*)wifiOutBuffer;
		packetHeader->audioIndex = sndWriterIndex + prevLocalAudioDelta;
		wifiOutBufferLength += sizeof(tPacketHeader);
}
		
	u32 cnt = 0;
	
	totalMicSamples += samplesToRead;
	
	while ( cnt < samplesToRead )
{
		// from mic to wifi out
		sample = micBufferMirror[micReaderPos];
		if( ++micReaderPos >= AUDIOBUFLEN) micReaderPos = 0;
							
		wifiOutBuffer[wifiOutBufferLength] = sample;
		samplesOut++;
		
		if (++wifiOutBufferLength >= AUDIOBUFLEN) 
			wifiOutBufferLength = AUDIOBUFLEN -1;
		
		cnt++;
}
	
	// insert samples to compensate for microphone lag caused by blocked interrupts
	if (localAudioDelta < 0)
{
		Talkie->micSampleCount++;
		IPC->mailData++;
		Talkie->micSampleCount++;
		IPC->mailData++;
		Talkie->micSampleCount++;
		IPC->mailData++;
		
		micCount++;
		micCount++;
		micCount++;
		
		wifiOutBuffer[wifiOutBufferLength++] = sample;
		if (wifiOutBufferLength >= AUDIOBUFLEN) wifiOutBufferLength = AUDIOBUFLEN -1;
		wifiOutBuffer[wifiOutBufferLength++] = sample;
		if (wifiOutBufferLength >= AUDIOBUFLEN) wifiOutBufferLength = AUDIOBUFLEN -1;
		wifiOutBuffer[wifiOutBufferLength++] = sample;
		if (wifiOutBufferLength >= AUDIOBUFLEN) wifiOutBufferLength = AUDIOBUFLEN -1;
		
		samplesOut++;
		samplesOut++;		
		samplesOut++;
}
	
		// flush buffer on contact timeout
	if (live == 0)
{
		samplesOut = 0;
		wifiOutBufferLength = 0;
}
	
	wifiInBufferPos = 0;
	u32 nextPacketPos;
	
	// get remote audio, possibly multiple packets
	while (wifiInBufferPos < wifiInBufferLength && live == 1)
{ 
		if ( (wifiInBufferLength - wifiInBufferPos) >= sizeof(tPacketHeader) )
{
			packetHeader = (tPacketHeader*)&wifiInBuffer[wifiInBufferPos];
			remotecounter = packetHeader->audioIndex;
			wifiInBufferPos += sizeof(tPacketHeader);
			nextPacketPos =  wifiInBufferPos + packetHeader->sampleCount;
			if (nextPacketPos > wifiInBufferLength)
				nextPacketPos = wifiInBufferLength;
			
			if (counterdelta == 0) // first frame
				counterdelta = sndWriterIndex - remotecounter;
			
			// where to place frame relative to own sndWriterIndex?
			sndWriterPos = (counterdelta + remotecounter) % AUDIOBUFLEN;
			if (sndWriterPos < 0) sndWriterPos += AUDIOBUFLEN;
			soundoffset = counterdelta + remotecounter - sndWriterIndex;
			
			// soundoffset < frames ahead: packet outside jitter buffer, advance to next packet
			if (soundoffset < -12 * audioBytesInFrame) wifiInBufferPos += nextPacketPos;
	
			// write input samples to sndBuffer
			while ( wifiInBufferPos < nextPacketPos )
{
				// from wifi in to sndBuffer
				sample = wifiInBuffer[wifiInBufferPos++];
				samplesIn++;
								
				sndBufferMirror[sndWriterPos] = sample ^ 0x80;
				if( ++sndWriterPos >= AUDIOBUFLEN) sndWriterPos = 0;
}	
}
		 else wifiInBufferPos = wifiInBufferLength;
}
	
	// clear part of played buffer
	sndWriterPos = ( sndIndex - (int)(10 * audioBytesInFrame) + 1 ) % AUDIOBUFLEN;
	if ( sndWriterPos >= -audioBytesInFrame ) 
{
		cnt = 0;
		while (cnt++ < audioBytesInFrame + 2) sndBufferMirror[sndWriterPos++] = 0;
}
		
	if (live == 0) // reset counter
		counterdelta = 0;
	
	if( test == -1 || test % 30 == 0 ) // 2x p sec 
		print_debug();

	wifiInBufferLength = 0;
	
	test++;
}

*/

void calc_power()
{	
		// calculate average power of signal, store max value in power display interval
		u32 sndIndex = get_mic_counterVB() - Talkie->sndOffset; // played sample counter
		u32 sndReaderPos = sndIndex % AUDIOBUFLEN;   // last played sample position in sndBuffer
	
		u32 ptr;
		
		u32 pwr = 0;
		
		ptr = sndReaderPos;
		u32 cnt = 0;
		int8 sample;
		int s;
		
		while ( cnt++ < 13 )
		{
			sample = sndBufferMirror[ptr];
			ptr += 10; if (ptr >= AUDIOBUFLEN) ptr = 0;
			s = sample * sample;
			pwr += s;
		}
		
		pwr = (u32)sqrt( pwr / 13 ); // max 128
		pwr = (u32)((75 * pwr) / 128);
		
		if (pwr > disppwr) 
			disppwr = pwr;
		
		if (disppwr > 75) disppwr = 75; // shouldn't happen
}


void userSwitched()
{
	// flush input
	wifiInBufferLength = 0;
	// restart sync
	counterdelta = 0;
}

void init()
{
	sndWriterPos = 0;
}


