/*

File: MessageQueue.c
Version: v0.2 nov 2008
Author: Eric Giesselbach

Drop-in replacement for liblobby's MessageQueue.
Using shared memory instead of IPC (increasing bandwidth)

Requires minor changes to libnds and liblobby, see the walkie-talkie README

*/

#include <nds.h>				// using REG_IME
#include "MessageQueue.h"
#include "smi_startup_viaipc.h"
#include "safe_malloc.h"
#include <string.h>
#include <malloc.h>
#include <smi_mutex.h>

typedef struct dataHeader 
{
	volatile unsigned long length;
	volatile unsigned long channel;
} __attribute__ ((__packed__)) tDataHeader;

volatile tDataHeader * volatile pDataHeader;

typedef struct bufferHeader 
{
	volatile unsigned long length;
	volatile unsigned long writePos;
} __attribute__ ((__packed__)) tBufferHeader;

volatile tBufferHeader * volatile inHeader;
volatile tBufferHeader * volatile outHeader;

volatile char * volatile inBuf;
volatile char * volatile outBuf;
volatile char * volatile ptr;

unsigned long len;
unsigned long *plen;
unsigned long readPos;
		
		
char *st_data;

SMI_MUTEX mutex;
LPSMI_MUTEX pMutex;
		
void	IPC_SendMessageForChannel(unsigned long channel,char *data,unsigned long length) ;

#ifdef ARM9
void armbufInit(unsigned long length)
{
	readPos = 0;
	
	pMutex = &mutex;	
	CreateSMIMutex(pMutex, false);
				
	outBuf = (char*)malloc(length + sizeof(tBufferHeader));
	inBuf  = (char*)malloc(length + sizeof(tBufferHeader));
	
	inBuf  = inBuf + 0x400000;
	outBuf = outBuf + 0x400000;
	
	st_data = (char*)malloc(length);
				
	inHeader  = (tBufferHeader*)inBuf;
	outHeader = (tBufferHeader*)outBuf;
	
	inHeader->length    = length;
	inHeader->writePos  = 0;
	outHeader->length   = length;
	outHeader->writePos = 0;
						
	IPC->toArm9buf = inBuf;
	IPC->toArm7buf = outBuf;
	
	inBuf  += sizeof(tBufferHeader);
	outBuf += sizeof(tBufferHeader);
	
	DC_FlushAll();
}
#else
void armbufInit()
{
	readPos = 0;
	
	pMutex = &mutex;	
	CreateSMIMutex(pMutex, false);
		
	inBuf = IPC->toArm7buf;
	outBuf = IPC->toArm9buf;

	inHeader  = (tBufferHeader*)inBuf;
	outHeader = (tBufferHeader*)outBuf;
	
	inBuf  += sizeof(tBufferHeader);
	outBuf += sizeof(tBufferHeader);
}	
#endif

void sleep(int vblanks)
{
	while (vblanks-- > 0)
		swiWaitForVBlank();
}
// returns false on buffer full
// requirement: data length < inHeader->length
bool sendData(unsigned long channel, char *data, unsigned long length)
{	
	
	if ( WaitForSMIMutex(pMutex, 30) != SMI_MUTEX_TIMEDOUT )
	{		
		int freeLen = outHeader->length - outHeader->writePos - sizeof(tDataHeader);
			
		if (freeLen < length)
		{
			ReleaseSMIMutex(pMutex);
			return false;
		}
			
		pDataHeader = (tDataHeader*)outBuf + outHeader->writePos;
		ptr = outBuf + outHeader->writePos + sizeof(tDataHeader);
	
		pDataHeader->length  = length;
		pDataHeader->channel = channel;
										
		memcpy((void*)ptr, data, length);
		outHeader->writePos += length + sizeof(tDataHeader);
		
		ReleaseSMIMutex(pMutex);
		
		return true;
	}
		else 
	{
		return false;
	}
}

// call this procedure to retrieve data packets until empty (returns false)
// requirement: data points to allocated memory, size >= length
bool receiveData(char *data, unsigned long *length)
{			

	if ( WaitForSMIMutex(pMutex, 30) != SMI_MUTEX_TIMEDOUT )
	{
		if (readPos >= inHeader->writePos)
		{
			ReleaseSMIMutex(pMutex);
			return false;
		}
			
		pDataHeader = (tDataHeader*)inBuf + readPos;
		
		if (pDataHeader->length >= 0)
		{
			// clip data on length
			if (*length > pDataHeader->length) *length = pDataHeader->length;
			
			ptr = inBuf + readPos + sizeof(tDataHeader);
			memcpy(data, (void*)ptr, *length);
			
			readPos += pDataHeader->length + sizeof(tDataHeader);
		}
		
		if (readPos >= inHeader->writePos)
		{
			readPos = 0;
			inHeader->writePos = 0;
		}
			
		ReleaseSMIMutex(pMutex);
		
		return true;
	}
	  else
	{
		return false;
	}
}

// call this procedure to fetch channel of first data packet
unsigned long senseData()
{			
	if (readPos >= inHeader->writePos)
		return 0;
	
	pDataHeader = (tDataHeader*)inBuf + readPos;
	
	return pDataHeader->channel;
}

bool dataWaiting()
{			
	return (readPos < inHeader->writePos);
}

typedef struct IPC_MESSAGEBUFFER
{
	unsigned char				* data ;
	unsigned long				length ;
	long						position ;
	struct IPC_MESSAGEBUFFER *	next ;
} IPC_MESSAGEBUFFER, *LPIPC_MESSAGEBUFFER ;

volatile LPIPC_MESSAGEBUFFER IPC_SendBuffers = 0 ;
volatile LPIPC_MESSAGEBUFFER IPC_RecvBuffers = 0 ;

IPC_MSGCOMPLETECALLBACK_PROC	msgCompleteCallback = 0 ;

void IPC_SetMsgCompleteCallback(IPC_MSGCOMPLETECALLBACK_PROC callback) 
{
	msgCompleteCallback = callback ;
}

typedef struct IPC_CHANNELCALLBACKS
{
	IPC_CHANNELCALLBACK				callback ;
	unsigned long					channel ;
	struct IPC_CHANNELCALLBACKS *	next ;
} IPC_CHANNELCALLBACKS, *LPIPC_CHANNELCALLBACKS ;

LPIPC_CHANNELCALLBACKS IPC_ChannelCallbacks = 0 ;

// not used in arm7 liblobby code
void IPC_SetChannelCallback(unsigned long channel, IPC_CHANNELCALLBACK cb)
{
	LPIPC_CHANNELCALLBACKS tmp = IPC_ChannelCallbacks ;
	while (tmp)
	{
		if (tmp->channel == channel)
		{
			tmp->callback = cb ;
			return ;
		}
		tmp = tmp->next ;
	}
	LPIPC_CHANNELCALLBACKS n = (LPIPC_CHANNELCALLBACKS)safe_malloc(sizeof(IPC_CHANNELCALLBACKS)) ;
	n->channel = channel ;
	n->callback = cb ;
		
	n->next = IPC_ChannelCallbacks ;
	IPC_ChannelCallbacks = n ;
}

volatile bool IPC_mallocErr = 0 ;
volatile int IPC_mallocErrSize = 0 ;

int IPC_LockSystem(int newLock)
{
	#ifdef WIN32
	#else
		int old = REG_IE ;
		REG_IE = newLock | IRQ_TIMER0;
		return old ;
	#endif
}

// used in arm9 part
// arm7 uses IPC_GetMessage
void IPC_CheckInput(void)
{
	unsigned long channel;
	
	channel = 0;
	
	if ( IPC->toArm9buf == 0 )
		return;
				
	int lock = IPC_LockSystem(0);
	
	len = 4000;
	plen = &len;
	
	LPIPC_CHANNELCALLBACKS tmp;
			
	while ( receiveData(st_data, plen) )
	{				
		tmp = IPC_ChannelCallbacks;
		
		while (tmp)
		{
			if (tmp->channel == channel)
			{
				IPC_LockSystem(lock);	
				tmp->callback(st_data, len) ;
				IPC_LockSystem(0);
			}
			
			tmp = tmp->next;
		}
		
		len = 4000;
	}
	
	IPC_LockSystem(lock);
}


void IPC_WaitForAllSent(void)
{
	//IPC_DrainFIFO();
}

void IPC_SendMessage(char *data,unsigned long length)
{
	IPC_SendMessageForChannel(0, data, length);
}

void IPC_SendMessageForChannel(unsigned long channel,char *data, unsigned long length)
{
	unsigned long savedIME = REG_IE ;
	REG_IE = IRQ_TIMER0 ;
	// we append from front and take from last
	
	sendData(channel, data, length);
	
	//IPC_DrainFIFO();
	
	REG_IE = savedIME ;
}

bool IPC_GetMessage(char *data, unsigned long *length)
{	
	int lock = IPC_LockSystem(0) ;
	
	bool gotMessage = receiveData(data, length);
	
	if (!gotMessage)
	{
		IPC_LockSystem(lock);
		return false;
	}

	IPC_LockSystem(lock) ;
	
	return true ;
}

int IPC_AllreadySending(unsigned char *data, int length)
{
	return 0 ;
}

int IPC_Init(void)
{
	unsigned long savedIME = REG_IE ;
	REG_IE = IRQ_TIMER0;
		
	armbufInit(22050);
	
	REG_IE = savedIME;
	return 1 ;
	
}

IPC_CUSTOMCALLBACK_PROC customIPCCallback = 0 ;

void IPC_SetCustomCallback(IPC_CUSTOMCALLBACK_PROC callback) 
{
	customIPCCallback = callback;
}

void IPC_PassCustomMessage(unsigned char *data, int length) 
{
	if (customIPCCallback)
		customIPCCallback(data, length) ;
}
